program l1radon
  !
  !     =======================================================
  !     PRIMAL DUAL LOG BARRIER LINEAR PROGRAMMING algorithm
  !     for the inversion of the Hyperbolic Radon Transform
  !     =======================================================
  !     Brad Artman
  !     Stanford Exploration Project
  !     Stanford University
  !
  !     M.D.Sacchi
  !     Signal Analysis and Imaging Group
  !     University of Alberta, Edmonton, Canada
  !   
  !     Notes
  !     1-Solve Lm-d with sparnseness contraints on model (vel) space
  !       using the principle of "Atomic Decomposition by Basis Pursuit"
  !       from Chen, Donoho, and Saunders: Technical Report 479, 
  !       Department of Statistics, Stanford University, May 1995.
  !       
  !     2-The algorithm is modified from standard LP to deal with 
  !       non-positive model space values
  !     3-A sparse matrix for HRT is built and applied rather than operators 
  !     4-Input cmp's
  !     5-Outputs:
  !        stdout        --> velocity panel (m),
  !        rayp=______.H --> ray parameter panel
  !        preddata=__.H --> modelled data (Lm), 
  !     6-Parameters:
  !       buildm - true/false iteratively writes model space
  !       theta0 - angle origin for ray param
  !       dtheta    - delta ray param
  !       ntheta    - number of ray params  
  !       v0 - velocity origin
  !       dv    - delta vel
  !       nv    - number of vel's
  !       tol_cg- tolerance to stop CG iterations
  !       itercg- number of CG iterations to perform
  !       iterlp- maximum number of LP iterations to perform w/o convergence
  !       delta - regularization of search direction for Interior Point LP method
  !               set to 1 for de-noising
  !       gamma - similar to delta, both should be ~10^-3, never 10^-4.
  !               more accuraccy, less stability at 10^-2
  !       mu    - scaling parameter (iteratively updated- ie. don't adjust it)
  !               for the logarithmic barrier sub-problem
  !       Convergence testing thresholds for Linear Prgramming problem
  !          pi- primal infeasability threshold
  !          di- daul infeasability threshold
  !          gap- duality gap measures closeness between primal & dual sol'ns


  use sep
  use hrt_builder
  implicit none

  real,   allocatable,dimension(:)  :: m,ma,c
  real,   allocatable,dimension(:)  :: d,pdat,de
  integer,allocatable,dimension(:)  :: irow,  icol,  a
  integer,allocatable,dimension(:)  :: t_irow,t_icol,t_a
  integer  :: nt,nh,nv,n3,nm,nd,na,nm2,na2,np,itercg,iterlp,i,i3
  real     :: dt,dh,t0,h0,dv,v0,d3,o3,dp,p0,delta
  real     :: mu,gamma,pi,di,dg,error,tol,circ
  integer  ::buildm
  logical  ::linear
  circ=acos(-1.)
  call sep_init()
  call from_history('n1',nt);call from_history('n2',nh);call from_history('n3',n3)
  call from_history('d1',dt);call from_history('d2',dh);call from_history('d3',d3)
  call from_history('o1',t0);call from_history('o2',h0);call from_history('o3',o3)

  call from_param('v0',v0,1400.);call from_param('dv',dv,25.);call from_param('nv',nv,60)

  call to_history('n1',nt);call to_history('n2',nv);call to_history('n3',n3)
  call to_history('d1',dt);call to_history('d2',dv);call to_history('d3',d3)
  call to_history('o1',t0);call to_history('o2',v0);call to_history('o3',o3)

  linear=.false.
  if (exist_file ("rayp")) then
     linear=.true.
     call from_param('theta0',p0,-0.000332);call from_param('ntheta',np,60);call from_param('dtheta',dp,-2*p0/np)
!     p0=-sin(p0*circ/180.)/(v0;     dp=2*p0/np
     call to_history('n1',nt,"rayp");call to_history('n2',np,"rayp");call to_history('n3',n3,"rayp")
     call to_history('d1',dt,"rayp");call to_history('d2',dp,"rayp");call to_history('d3',d3,"rayp")
     call to_history('o1',t0,"rayp");call to_history('o2',p0,"rayp");call to_history('o3',o3,"rayp")    
  end if

  if (exist_file ("preddata")) then
     call to_history('n1',nt,"preddata");call to_history('n2',nh,"preddata");call to_history('n3',n3,"preddata")
     call to_history('d1',dt,"preddata");call to_history('d2',dh,"preddata");call to_history('d3',d3,"preddata")
     call to_history('o1',t0,"preddata");call to_history('o2',h0,"preddata");call to_history('o3',o3,"preddata")
  end if

  call from_param('iterlp',iterlp,30)
  call from_param('itercg',itercg,nv*nt/10)
  call from_param('tol_cg',tol, 1.e-1)
  call from_param('delta',delta,1.e-0)   
  call from_param('gamma',gamma,1.e-3) 
  call from_param('gap',dg,1.e-1)
  call from_param('pi',pi,dg)
  call from_param('di',di,dg)
  call from_param('mu',mu,5.e-2)
  call from_param('buildm',buildm,0)

  call sep_close()

  nm=nv*nt     ! size of the vector of unknowns
  if (linear) nm=nv*nt+nt*np ! size of the vector of unknowns
  nd=nh*nt     ! size of the vector containing the data 

  !! allocate temp arrays for HRT operator
  if (linear) then
     allocate(t_irow(nt*nh*nv+nt*nh*np),t_icol(nt*nh*nv+nt*nh*np),t_a(nt*nh*nv+nt*nh*np))
  else
     allocate(t_irow(nt*nh*nv),t_icol(nt*nh*nv),t_a(nt*nh*nv))
  end if
  write(0,*) ' '
  write(0,*) " Calculating Hyperbolic Radon sparse matrix"
  if (linear)then
     call HRTmat(nh,dh,h0,nt,dt,t0,nv,dv,v0,t_icol,t_irow,t_a,na,linear,p0,dp,np)
  else
     call HRTmat(nh,dh,h0,nt,dt,t0,nv,dv,v0,t_icol,t_irow,t_a,na,linear)
  end if

  !!  Modify problem to handle non-positive solutions.
  !!  This is to expand the operator as [L,-L] [u,v]^T = d where 
  !!  the solution is u,-v with positivity on u and v.
  !!  I'm using ma=[u,v] rather than introducing new variables u and v
  !!  [ma for model augmented]
  na2 = 2*na 
  nm2 = 2*nm

  !!allocate real HRT sparse matrix arrays and fill
  allocate(irow(na2),icol(na2),a(na2))
  irow(1:na)=t_irow(1:na)
  icol(1:na)=t_icol(1:na)
  a(1:na)=   t_a(1:na)
  deallocate(t_irow,t_icol,t_a)

  irow(na+1:na2)=irow(1:na)
  icol(na+1:na2)=icol(1:na)+nm
  a(na+1:na2)=-a(1:na)

  allocate(ma(nm2),c(nm2));ma=1. ;c=1. !!ma as above, c from Chen et al.
  write(0,*) ' '
  write(0,*)' -------------------------------------------'  
  write(0,*)' Number of CMPs              : ', n3
  write(0,*)' Number of traces            : ', nh 
  write(0,*)' Number of velocities        : ', nv 
  write(0,*)' Minimum velocity            :  ', v0
  write(0,*)' Maximum velocity            :  ', v0+(nv-1)*dv
  if (linear) then
  write(0,*)' Number of ray parameters    : ', np 
  write(0,*)' Minimum ray parameter       :  ',p0
  write(0,*)' Maximum ray parameter       :  ',p0+(np-1)*dp
  end if
  write(0,*)' Number of t and tau samples : ', nt
  write(0,*) 
  write(0,*)' Number of data points       : ', nd
  write(0,*)' Number of model parametes   : 2*', nm
  write(0,'(a,i6,a,i6)')'  Size of the Non-Sparse Op.  :', nd,' x',nm
  write(0,*)' Number of non-zero elements : ', na
  write(0,'(a,f5.2,a)')'  Sparsity of HRT matrix      :        ', 100.*na/(nm*nd),'%'
  write(0,*) 
  write(0,*)'    PDLBBP Parameters'
  write(0,*)' delta                       :',delta
  write(0,*)' gamma                       :',gamma
  write(0,*)' gap tolerance               :',dg
  write(0,*)' Maximum BP iterartions      :',iterlp
  write(0,*)' Maximum CG iterartions      :',itercg
  if(buildm==1) write(0,*) 'Iterative model write in the shiz-ouse '
  allocate(d(nd));d=0.

  do i3=1, n3
     write(0,*)'-----> Working on CMP number',i3
     call sep_read(d)
     !!do the deed: ma is the output, c is the cost function, d is the force vector
     if(buildm) then
        allocate(m(nm2))
        call PD_lp_bm(nm2,nd,a,na2,c,ma,d,irow,icol,iterlp,itercg,tol,delta,gamma,mu,PI,DI,DG,buildm,linear,m,nv,np,nt)
        deallocate(m)
     else
        call PD_lp_bm(nm2,nd,a,na2,c,ma,d,irow,icol,iterlp,itercg,tol,delta,gamma,mu,PI,DI,DG,buildm)
     end if
     !!  Some figures of merit to show how we did
     write(0,*)' PDLPBM Iterations    :', iterlp
     write(0,*)' Primal Infeasibility :', PI
     write(0,*)' Dual Infeasibility   :', DI
     write(0,*)' Dualilty Gap         :', DG

     !!  ma is strickly positive- form the non strickly positive variable n
     allocate(m(nm))
     do i=1,nm
        m(i) = ma(i)-ma(i+nm)
     enddo
     deallocate(ma)
     if (linear)then
        call sep_write(m(1:nv*nt))
        call sep_write(m(nv*nt+1:nv*nt+np*nt),"rayp")
     else
        call sep_write(m)
     end if

     !!  forward model to obtain the predicted data 
     allocate(pdat(nd),de(nd))
     call mat_mul (0,a,na,nm,m,nd,pdat,irow,icol)
     if (exist_file ("preddata")) then
        call sep_write(pdat,'preddata')
     end if

     deallocate(m)
     de=d-pdat
     error=sqrt(dot(nd,de,de)/nd)
     deallocate(pdat,d,de)

     write(0,*)' MSE                  :', error
     write(0,*)' -----------------------------------------'  
     write(0,*)
!     call to_history("BPiters",iterlp)
!     if (exist_file ("preddata")) call to_history("BPiters",iterlp,"preddata")
!     if (exist_file ("rayp"))     call to_history("BPiters",iterlp,"rayp")
  end do
  call exit (0) 
end program l1radon
