module norm_mod 
 use sep_mod
  implicit none

  type norm_obj
     private
     real ::rt  ! rt = r_thr = r_bar for some norms 
     character(10)   ::norm_type ! Type of norm to be used for plane search
  end type norm_obj
contains

  !Private functions
  ! ========================================================================================
  !                   NORM Functions 
  ! ========================================================================================
  !!  Arguments IN: 
  !!     rr : residual (R)
  !!  Arguments OUT: 
  !!     c0 : Norm value at R                 (C )
  !!     c1 : Norm 1st derivative value at R  (C')
  !!     c2 : Norm 2nd derivative value at R  (C")

  !!  Output Flag : 
  !!     1 : Exit without completion
  !!     0 : Done without any exit or error   
  !!

  ! --------------------------------------------
  !       SIGNUM (L1) norm and Derivative
  ! --------------------------------------------
  integer function l1( rr, c0,c1,c2)
    !  Computes SIGNUM(L1) Norm and its 1st,2nd derivative around point rr
    real, dimension (:)            :: rr
    real, dimension(:) :: c0,c1,c2
    integer                        :: i

    l1=1
    do i=1,size(rr)

       c0(i)= abs(rr(i)) 
       c1(i)= rr(i)/max(abs(rr(i)),1.d-12)
       ! Implements the dirac delta function
       c2(i) = 2/(1e-6*sqrt(3.14159)) * exp( - (rr(i))**2 * 1e12 )  
    end do

    l1=0
  end function l1



  ! --------------------------------------------
  !       CLASSIC (L2) norm and Derivative
  ! --------------------------------------------
  integer function l2( rr, c0,c1,c2)
    !  Computes L2 (classic) Norm and its 1st,2nd derivative around point rr
    real, dimension (:)            :: rr
    real, dimension(:) :: c0,c1,c2
    integer                        :: i

    l2=1
    !$OMP PARALLEL DO private(i)
    do i=1,size(rr)
       c0(i)= rr(i)*rr(i)/2
       c1(i)= rr(i)
       c2(i)= 1
    end do
    !$OMP END PARALLEL DO
    l2=0
  end function l2

  ! --------------------------------------------
  !         HYBRID norm and Derivative
  ! --------------------------------------------
  integer function hybrid( rt, rr, c0,c1,c2)
    !  Computes HYBRID Norm and its 1st,2nd derivative around point rr
    ! C(0) = 0
    real :: rt
    real, dimension (:)            :: rr
    real, dimension(:) :: c0,c1,c2
    double precision               :: term
    integer                        :: i

    !!Change the definition of C=R^2*(...) to C=R*(....)
    !!such that when R->0, it tends to C=|r| without a scaling factor R
    hybrid=1
    !$OMP PARALLEL DO private(i)
    do i=1,size(rr)
       !! Jon's HYBRID
       term = sqrt(1+rr(i)**2/(rt**2))
       c0(i)= rt**1 * (term -1) 
       c1(i)= rr(i)/(rt*term)
       c2(i)= 1./(term**3*rt)
    end do
    !$OMP END PARALLEL DO
    hybrid=0
  end function hybrid


  ! --------------------------------------------
  !         HUBER norm and Derivative
  ! --------------------------------------------
  integer function huber( rt, rr, c0,c1,c2)
    !  Computes HUBER Norm and its 1st,2nd derivative around point rr
    !  Note that rt > 0 (set in init. function)
    real, dimension (:)            :: rr
    real ::rt
    real, dimension(:) :: c0,c1,c2
    double precision               :: ratio
    integer                        :: i

    huber=1
    do i=1,size(rr)
       ratio = rr(i)/rt

       ! Jon's HUBER
       if (abs(ratio)>=1.) then
          c0(i)= abs(rr(i))-rt/2. 
          c1(i)= rr(i) / max(abs(rr(i)),1.d-12)
          c2(i) = 0.              ! Huber linear portion has 0 OR Inf. convexity 

       else
          c0(i)= rt* ratio**2/2.
          c1(i)= ratio
          c2(i)= 1/rt
       end if

    end do

    huber=0
  end function huber




  !Public functions

  ! --------------------------------------------
  !    Compute the norm C(r)
  ! --------------------------------------------
  integer function norm_c0(S_norm_in,rd,c_r)
    ! Compute g for different norms (rd,norm,rthr --> rw)
    !
    !!  Arguments IN: 
    !!     rd   : -Wd +WFm0
    !!     S_norm: The norm object
    !!  Arguments OUT: 
    !!     c_r   : C(r), the norm 

    !!  Output Flag : 
    !!     1 : Exit without completion
    !!     0 : Done without any exit or error   
    !!
    type(norm_obj) :: S_norm_in

    real, dimension(:)  :: rd
    integer ::n_rd

    real, dimension(:)  :: c_r

    real, dimension(:), allocatable     ::tmp0,tmp1,tmp2

    n_rd = size(rd)
    allocate (tmp0(n_rd),tmp1(n_rd),tmp2(n_rd)) 
    norm_c0 = 1
    norm_c0 = compute_norm(S_norm_in,rd,tmp0,tmp1,tmp2)
    c_r = tmp0
    deallocate(tmp0,tmp1,tmp2)

  end function norm_c0

  ! --------------------------------------------
  !    Derivative step for different norms  C'(r)
  ! --------------------------------------------
  integer function norm_c1(S_norm_in,rd,rw)
    ! Compute g for different norms (rd,norm,rthr --> rw)
    !
    !!  Arguments IN: 
    !!     rd   : -Wd +WFm0
    !!     norm_obj: S_norm_in
    !!  Arguments OUT: 
    !!     rw   : C'(norm)*rd 

    !!  Output Flag : 
    !!     1 : Exit without completion
    !!     0 : Done without any exit or error   
    
    type (norm_obj)  S_norm_in
    real, dimension(:)  :: rd
    integer ::n_rd

    real, dimension(:)  :: rw

    real, dimension(:), allocatable     :: tmp0,tmp1,tmp2

    n_rd = size(rd)
    allocate (tmp0(n_rd),tmp1(n_rd),tmp2(n_rd)) 

    norm_c1=1
    norm_c1 = compute_norm(S_norm_in,rd,tmp0,tmp1,tmp2)
    rw = tmp1
    deallocate(tmp0,tmp1,tmp2)
    norm_c1=0
  end function norm_c1

  ! --------------------------------------------
  !        Compute the modulus of a vector, eg sum(C(r))
  ! --------------------------------------------

  integer function norm_modulus(S_norm_in,r_in,A_out)
    type (norm_obj)  S_norm_in
    real, dimension(:)  :: r_in
    real   :: A_out

    integer ::n_rd
    real, dimension(:), allocatable     :: tmp0

    n_rd = size(r_in)
    allocate (tmp0(n_rd))

    norm_modulus=1
    norm_modulus = norm_c0(S_norm_in,r_in,tmp0)
    
    A_out = sum(tmp0)

    deallocate(tmp0)
    norm_modulus=0
  end function norm_modulus

  ! --------------------------------------------
  !         Initialize norm object
  ! --------------------------------------------
  subroutine norm_init(S_norm_in,rthr,norm_in)
    ! Initialize CGNORM Stepper with r_thr, norm type and psiter_in
    !
    ! Arguments IN: 
    !    rthr      : RBAR or R_THR for norm 
    !    norm_in   : Norm ype to be used (Hybrid or Huber)
    !    S_norm_in : norm structure
    type(norm_obj)  S_norm_in
    real           :: rthr
    character(12)  :: norm_in
    real        ::rt

    rt = abs(rthr)     ! Rt is positive
    if (norm_in(1:5)=="huber" .or. norm_in(1:6)=="hybrid") then
       if (rt<=0.0)  then 
          write(0,*) "r_thr is ", rt
          call seperr('r_thr <=0 is not a valid choice.')
       endif
    else if ((norm_in(1:2)=="l1" .or. norm_in(1:2)=='l2')) then
       rt = 0
    else
       write(0,*) "norm specification error, unknown type of norm ",norm_in
       call seperr("norm  specification error, unknown type of norm ")
       return
    end if

    S_norm_in%norm_type = norm_in
    S_norm_in%rt = rt

    write(0,*) "The norm we choose is ",S_norm_in%norm_type, "rt",rt
  end subroutine norm_init

  ! --------------------------------------------
  !         Check whether the type of norm needs Threshold
  ! --------------------------------------------
  function norm_need_Rth( str_norm_in) result(need_Rth)
    logical :: need_Rth
    character(12) ::str_norm_in
    need_Rth = .false.
    if (str_norm_in(1:5)=="huber" .or. str_norm_in(1:6)=="hybrid") then
       need_Rth = .true.
    else if ((str_norm_in(1:2)=="l1" .or. str_norm_in(1:2)=='l2')) then
       need_Rth = .false.
    else
       write(0,*) "norm specification error, unknown type of norm ",str_norm_in
       call seperr("norm  specification error, unknown type of norm ")
       return
    end if

  end function norm_need_Rth
  
  integer function compute_norm( S_norm_in, rr, c0,c1,c2)
    !!  Output Flag : 
    !!     1 : error occurs
    !!     0 : Done without any exit or error  
    type(norm_obj) S_norm_in
    real, dimension (:)            :: rr
    real, dimension(:) :: c0,c1,c2

    character(12) :: norm
    real ::rt
    compute_norm = 0
    norm = S_norm_in%norm_type
    rt = S_norm_in%rt

    if      (norm(1:2)=="l2")     then
       compute_norm = l2     (rr, c0,c1,c2)
    else if (norm(1:2)=="l1")     then
       compute_norm = l1     (rr, c0,c1,c2)
    else if (norm(1:6)=="hybrid")  then
       compute_norm = hybrid (rt,rr, c0,c1,c2)
    else if (norm(1:5)=='huber')   then
       compute_norm = huber  (rt,rr, c0,c1,c2)
    end if

  end function compute_norm

  ! --------------------------------------------
  !             Compute the average r given a average penalty function (measuring the Fitting success)
  ! --------------------------------------------
  integer function compute_avg_r_from_c(S_norm_in, c_in, value)
    !!  Output Flag : 
    !!     1 : error occurs
    !!     0 : Done without any exit or error  
    !!    Input arg: c_in,norm_obj
    !!    Output arg: value

    type(norm_obj) S_norm_in
    real ::c_in
    real, intent(out) ::value

    character(12) ::str_norm
    real :: rt

    compute_avg_r_from_c = 0
    str_norm = S_norm_in%norm_type
    rt = S_norm_in%rt

    value = 0.0
    if (str_norm(1:2)=="l2")     then
       value = sqrt(c_in)
    else if (str_norm(1:2)=="l1")     then
       value = c_in
    else if (str_norm(1:6)=="hybrid")  then
       value = sqrt((c_in+rt)**2 - rt**2)
    else if (str_norm(1:5)=='huber')   then
       value = c_in
       if (value >= rt*0.5) then
          value = value + rt*0.5
       else
          value = sqrt(value*2*rt)
       endif
    endif
    compute_avg_r_from_c=1
end function compute_avg_r_from_c

! --------------------------------------------
!             Compute the average r given a residue vector (measuring the Fitting success)
! --------------------------------------------
integer function compute_avg_r(S_norm_in, rr, value)
  !!  Output Flag : 
  !!     1 : error occurs
  !!     0 : Done without any exit or error  
  !!    Input arg: rr,norm_obj
  !!    Output arg: value

  type(norm_obj) S_norm_in

  real, dimension(:) ::rr
  real, intent(out) ::value
  character(12) ::str_norm
  integer :: N
  real :: rt
  real :: avg_c
  real, dimension(:), allocatable ::c0,c1,c2

  compute_avg_r = 0
  str_norm = S_norm_in%norm_type
  rt = S_norm_in%rt

  value = 0.0
  N = size(rr)
  allocate(c0(N),c1(N),c2(N))

  if (str_norm(1:2)=="l2")     then
     compute_avg_r  = l2     (rr, c0,c1,c2)
  else if (str_norm(1:2)=="l1")     then
     compute_avg_r = l1     (rr, c0,c1,c2)
  else if (str_norm(1:6)=="hybrid")  then
     compute_avg_r = hybrid (rt,rr, c0,c1,c2)
  else if (str_norm(1:5)=='huber')   then
     compute_avg_r = huber  (rt,rr, c0,c1,c2)
  endif

  avg_c = sum(c0)/N
  compute_avg_r = compute_avg_r_from_c(S_norm_in,avg_c,value)
  deallocate(c0,c1,c2)
end function compute_avg_r

  !-----------------------------------------------
  !               Quantile function: Find the kth smallest value from array a
  !		eg. if k=1		 : Find the smallest value, Zero spikeness
  !			k=size(a)	: Find the biggest value, Full spikenss
  !-----------------------------------------------
  recursive function quantile(k,a) result( value)
    integer,             intent (in)  :: k          
    ! position in array
    real, dimension (:), intent (in)  :: a
    real                              :: value      
    ! output value of quantile
    integer                           :: j
    real                              :: ak
    ak = a(k)
    j = count( a < ak)  
    ! how many a(:) < ak
    if ( j >= k) then
       value = quantile( k, pack( a, a < ak))
    else
       j = count( a > ak) + k - size( a)
       if ( j > 0) then
          value = quantile( j, pack( a, a > ak))
       else
          value = ak
       end if
    end if
  end function

end module norm_mod
